import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.Arrays;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.KeyStroke;

/***
 * This class is the one to extend if you want to implement some ruleset into the presentation.
 * All you need to do is implement the nextstep method with whatever changes you want to make to the cellspace array
 * every generation.
 * Here is where code specific to rendering and navigating cellspaces is kept.
 */
public abstract class CellularSlide extends BasicSlide {

	private static final long serialVersionUID = 1L;
	
	int width, height, cellsize, padding, leftspace, upspace, time;
	
	boolean singlestep = true, backwards, tracking;
	
	ArrayList<int[][]> history;
	
	protected int[][] cellspace;
	protected Color[] colortable;

	public CellularSlide(SlideShow s, CellSpace c)
	{
		super(s);
		
		width = c.width;
		height = c.height;
		padding = c.padding;
		colortable = c.colortable;
		tracking = c.tracking;
		
		history = new ArrayList<int[][]>();
		
		this.cellsize =  Math.min((s.ysize-padding)/height, (s.xsize-padding)/width) - padding;
		
		leftspace = ( s.xsize -(width*(cellsize+padding) - padding))/2;
		upspace = (   s.ysize-(height*(cellsize+padding) - padding))/2;
		
		cellspace = new int[height][];
		for(int i = 0; i < height; i++)
		{
			cellspace[i] = new int[width];
		}
		
		Action slower = new CellularAction(CellularAction.SLOWER);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("F9"), "slower");
		this.getActionMap().put("slower", slower);
		
		Action faster = new CellularAction(CellularAction.FASTER);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("F10"), "faster");
		this.getActionMap().put("faster", faster);
		
		Action normal = new CellularAction(CellularAction.NORMAL);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("F11"), "normal");
		this.getActionMap().put("normal", normal);
		
		Action stepping = new CellularAction(CellularAction.STEPPING);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("F12"), "stepping");
		this.getActionMap().put("stepping", stepping);
		
		Action backwards = new CellularAction(CellularAction.BACKWARDS);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("F7"), "backwards");
		this.getActionMap().put("backwards", backwards);
		
		Action forwards = new CellularAction(CellularAction.FORWARDS);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("F8"), "forwards");
		this.getActionMap().put("forwards", forwards);
		
		Action stepback = new CellularAction(CellularAction.STEPBACK);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("LEFT"), "stepback");
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("UP"), "stepback");
		this.getActionMap().put("stepback", stepback);
		
		Action stepon = new CellularAction(CellularAction.STEPON);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("RIGHT"), "stepon");
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("DOWN"), "stepon");
		this.getActionMap().put("stepon", stepon);
		
		Action reset = new CellularAction(CellularAction.RESET);
		this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("F8"), "reset");
		this.getActionMap().put("reset", reset);


		initCellspace();
		
		int[][] copy = new int[cellspace.length][];
		for (int i = 0; i < copy.length; i++) copy[i] = Arrays.copyOf(cellspace[i], cellspace[i].length);
		history.add(copy);
	}
	
	@Override
	public void render(Graphics g) {
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++){
				g.setColor(colortable[cellspace[y][x]]);
				g.fillRect(leftspace + x*(padding+cellsize), upspace + y*(padding + cellsize), cellsize, cellsize);
			}
		}
	}

	public abstract void nextstep();
	public abstract void initCellspace();
	
	public synchronized void cycle(){
		
		if (!singlestep){
			if (!backwards){
				time++;
			} else {
				if (time > 0)time--;
			}
		}
		
		if (time == history.size()){
			
			nextstep();
			
			int[][] copy = new int[cellspace.length][];
			for (int i = 0; i < copy.length; i++) copy[i] = Arrays.copyOf(cellspace[i],cellspace[i].length);
			history.add(copy);
			if(!tracking && time > 0) history.set(time - 1, null); // kill past if no tracking
			System.out.println("calculating:" + time);
		}
		
		cellspace = history.get(time);
		
		System.out.println("loading:" + time);
		
		for(int i = history.size() - 1; i > time; i--) history.remove(i); // kill future so that it gets recalculated when needed, to save ram
		
		changed = true; // this is not optimal
	}
	
	class CellularAction extends AbstractAction {
		static final int SLOWER = 0;
		static final int FASTER = 1;
		static final int NORMAL = 2;
		static final int RESET = 3;
		static final int STEPPING = 4;
		static final int FORWARDS = 5;
		static final int BACKWARDS = 6;
		static final int STEPON = 7;
		static final int STEPBACK = 8;
		private static final long serialVersionUID = 1L;
		
		int type;
		
		public CellularAction(int t){
			type = t;
		}
		
		public void actionPerformed(ActionEvent e) {
			switch (type){
			case CellularAction.SLOWER:
				if (rate == 0) rate = 1;
				else if (rate <= defaultrate*16) rate *= 2;
				break;
			case CellularAction.FASTER:
				rate = rate / 2;
				break;
			case CellularAction.NORMAL:
				rate = defaultrate;
				break;
			case CellularAction.RESET:
				if(singlestep){
					time=0;
					history.clear();
					cellspace = new int[cellspace.length][cellspace[0].length];
					initCellspace();
					int[][] copy = new int[cellspace.length][];
					for (int i = 0; i < copy.length; i++) copy[i] = Arrays.copyOf(cellspace[i], cellspace[i].length);
					history.add(copy);
					rate = defaultrate;
					changed = true;
				}
				break;
			case CellularAction.STEPPING:
				singlestep = !singlestep;
				break;
			case CellularAction.BACKWARDS:
				backwards = true;
				break;
			case CellularAction.FORWARDS:
				backwards = false;
				break;
			case CellularAction.STEPON:
				if (singlestep){
					time++;
					cycle();
				}
				break;
			case CellularAction.STEPBACK:
				if (tracking && singlestep && time > 0){
					time--;
					cycle();
				}
				break;
			}
		}
	}
}
